#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This test shows a surrogate-assisted Bayesian calibration of a time dependent
    analytical function.

Author: Farid Mohammadi, M.Sc.
E-Mail: farid.mohammadi@iws.uni-stuttgart.de
Department of Hydromechanics and Modelling of Hydrosystems (LH2)
Institute for Modelling Hydraulic and Environmental Systems (IWS), University
of Stuttgart, www.iws.uni-stuttgart.de/lh2/
Pfaffenwaldring 61
70569 Stuttgart

Created on Fri Aug 9 2019

"""
import numpy as np
import pandas as pd
import scipy.io as io
import sys
import joblib
# import bayesvalidrox
# Add BayesValidRox path
sys.path.append("../../src/")

from bayesvalidrox.pylink.pylink import PyLinkForwardModel
from bayesvalidrox.surrogate_models.inputs import Input
from bayesvalidrox.surrogate_models.engine import Engine
from bayesvalidrox.surrogate_models.exp_designs import ExpDesigns
from bayesvalidrox.bayes_inference.bayes_inference import BayesInference
from bayesvalidrox.bayes_inference.discrepancy import Discrepancy
from bayesvalidrox.bayes_inference.bayes_model_comparison import BayesModelComparison
import matplotlib
matplotlib.use('agg')


if __name__ == "__main__":

    # Number of parameters
    ndim = 10  # 2, 10

    # =====================================================
    # =============   COMPUTATIONAL MODEL  ================
    # =====================================================
    Model = PyLinkForwardModel()

    Model.link_type = 'Function'
    Model.py_file = 'analytical_function'
    Model.name = 'AnalyticFunc'

    Model.Output.names = ['Z']

    # For Bayesian inversion synthetic data with X=[0,0]
    Model.observations = {}
    Model.observations['Time [s]'] = np.arange(0, 10, 1.) / 9
    Model.observations['Z'] = np.repeat([2.], 10)

    # For Checking with the MonteCarlo refrence
    Model.mc_reference = {}
    Model.mc_reference['Time [s]'] = np.arange(0, 10, 1.) / 9
    Model.mc_reference['mean'] = np.load(f"data/mean_{ndim}.npy")
    Model.mc_reference['std'] = np.load(f"data/std_{ndim}.npy")

    # =====================================================
    # =========   PROBABILISTIC INPUT MODEL  ==============
    # =====================================================
    # Define the uncertain parameters with their mean and
    # standard deviation
    Inputs = Input()

    for i in range(ndim):
        Inputs.add_marginals()
        Inputs.Marginals[i].name = "$\\theta_{"+str(i+1)+"}$"
        Inputs.Marginals[i].dist_type = 'uniform'
        Inputs.Marginals[i].parameters = [-5, 5]

    # ------------------------------------------------
    # ------ Experimental Design Configuration -------
    # ------------------------------------------------
    ExpDesign = ExpDesigns(Inputs)
    
    # One-shot (normal) or Sequential Adaptive (sequential) Design
    ExpDesign.method = 'normal'
    ExpDesign.n_init_samples = 3*ndim

    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley
    # 6) chebyshev(FT) 7) grid(FT) 8)user
    ExpDesign.sampling_method = 'latin_hypercube'
    
    # Provide the experimental design object with a hdf5 file
    # ExpDesign.hdf5_file = 'ExpDesign_AnalyticFunc.hdf5'

    # Defining the measurement error, if it's known a priori
    obsData = pd.DataFrame(Model.observations, columns=Model.Output.names)
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = obsData**2

    # Create the engine
    engine = Engine(None, Model, ExpDesign)
    engine.train_normal()
    
    # =====================================================
    # ========  PostProcessing on the Model only ==========
    # =====================================================
    
    # So far there is no postprocessing for the model without surrogates!
    
    
    # =====================================================
    # ========  Bayesian inference without Emulator ==========
    # =====================================================
    BayesOpts = BayesInference(engine)
    BayesOpts.emulator = False
    BayesOpts.plot_post_pred = True

    # Select the inference method
    import emcee
    BayesOpts.inference_method = "MCMC"
    # Set the MCMC parameters passed to self.mcmc_params
    BayesOpts.mcmc_params = {
        'n_steps': 1e3,
        'n_walkers': 30,
        'moves': emcee.moves.KDEMove(),
        'multiprocessing': False,
        'verbose': False
        }

    # ----- Define the discrepancy model -------
    BayesOpts.measurement_error = obsData
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = obsData**2
    BayesOpts.Discrepancy = DiscrepancyOpts

    # Start the calibration/inference
    Bayes_PCE = BayesOpts.create_inference()

    # Save class objects
    with open(f'Bayes_{Model.name}.pkl', 'wb') as output:
        joblib.dump(Bayes_PCE, output, 2)

    # =====================================================
    # ========  Model Comparison without Emulator ==========
    # =====================================================
    sigma = 0.6
    data = {
        'x [m]': np.linspace(0.25, 4.75, 15),
        'Z': io.loadmat('data/synth_data.mat')['d'].reshape(1, -1)[0]
        }

    # =====================================================
    # ============   COMPUTATIONAL MODELS  ================
    # =====================================================
    # Define the models options
    # ---------- Linear model  -------------
    myL2Model = PyLinkForwardModel()

    myL2Model.link_type = 'Function'
    myL2Model.py_file = 'L2_model'
    myL2Model.name = 'linear'
    myL2Model.Output.names = ['Z']
    myL2Model.observations = data
    myL2Model.store = False

    # -- Nonlinear exponential model -------
    myNL2Model = PyLinkForwardModel()

    myNL2Model.link_type = 'Function'
    myNL2Model.py_file = 'NL2_model'
    myNL2Model.name = 'exponential'
    myNL2Model.Output.names = ['Z']
    myNL2Model.observations = data
    myNL2Model.store = False

    # ------ Nonlinear cosine model ---------
    # Data generating process
    myNL4Model = PyLinkForwardModel()

    myNL4Model.link_type = 'Function'
    myNL4Model.py_file = 'NL4_model'
    myNL4Model.name = 'cosine'
    myNL4Model.Output.names = ['Z']
    myNL4Model.observations = data
    myNL4Model.store = False

    # =====================================================
    # =========   PROBABILISTIC INPUT MODEL  ==============
    # =====================================================
    # Define model inputs
    n_sample = 10000
    # ---------- Linear model  -------------
    L2_Inputs = Input()
    L2_prior_mean = np.array([1, 0])
    L2_prior_cov = np.array(
        [[0.04, -0.007],
         [-0.007, 0.04]]
        )
    L2_input_params = np.random.multivariate_normal(
        L2_prior_mean, L2_prior_cov, size=n_sample
        )

    for i in range(L2_input_params.shape[1]):
        L2_Inputs.add_marginals()
        L2_Inputs.Marginals[i].name = f'$X_{i+1}$'
        L2_Inputs.Marginals[i].input_data = L2_input_params[:, i]

    # ------ Nonlinear exponential model ---------
    NL2_Inputs = Input()
    NL2_prior_mean = np.array([0.4, -0.3])
    NL2_prior_cov = np.array(
        [[0.003, -0.0001],
         [-0.0001, 0.03]]
        )
    NL2_input_params = np.random.multivariate_normal(
        NL2_prior_mean, NL2_prior_cov, size=n_sample
        )

    for i in range(NL2_input_params.shape[1]):
        NL2_Inputs.add_marginals()
        NL2_Inputs.Marginals[i].name = f'$X_{i+1}$'
        NL2_Inputs.Marginals[i].input_data = NL2_input_params[:, i]

    # ------ Nonlinear cosine model ---------
    NL4_Inputs = Input()
    NL4_prior_mean = np.array([2.6, 0.5, -2.8, 2.3])
    NL4_prior_cov = np.array(
        [[0.46, -0.07, 0.24, -0.14],
         [-0.07, 0.04, -0.05, 0.02],
         [0.24, -0.05, 0.30, -0.16],
         [-0.14, 0.02, -0.16, 0.30]]
        )
    NL4_input_params = np.random.multivariate_normal(
        NL4_prior_mean, NL4_prior_cov, size=n_sample
        )

    for i in range(NL4_input_params.shape[1]):
        NL4_Inputs.add_marginals()
        NL4_Inputs.Marginals[i].name = f'$X_{i+1}$'
        NL4_Inputs.Marginals[i].input_data = NL4_input_params[:, i]

    # ------------------------------------------------
    # ------ Experimental Design Configuration -------
    # ------------------------------------------------
    #L2_MetaModelOpts.add_ExpDesign()
    L2_ExpDesign = ExpDesigns(L2_Inputs)

    # One-shot (normal) or Sequential Adaptive (sequential) Design
    L2_ExpDesign.n_init_samples = 100

    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley
    # 6) chebyshev(FT) 7) grid(FT) 8)user
    L2_ExpDesign.sampling_method = 'latin_hypercube'

    # ------ Nonlinear cosine model ---------
    NL2_ExpDesign = ExpDesigns(NL2_Inputs)
    NL2_ExpDesign.method = 'normal'
    NL2_ExpDesign.n_init_samples = 100
    NL2_ExpDesign.sampling_method = 'latin_hypercube'

    # ------ Nonlinear cosine model ---------
    NL4_ExpDesign = ExpDesigns(NL4_Inputs)
    NL4_ExpDesign.method = 'normal'
    NL4_ExpDesign.n_init_samples = 100
    NL4_ExpDesign.sampling_method = 'latin_hypercube'

    # >>>>>> Train the Surrogates <<<<<<<<<<<
    L2_engine = Engine(None, myL2Model, L2_ExpDesign)
    L2_engine.train_normal()
    NL2_engine = Engine(None, myNL2Model, NL2_ExpDesign)
    NL2_engine.train_normal()
    NL4_engine = Engine(None, myNL4Model, NL4_ExpDesign)
    NL4_engine.train_normal()
    
    # =====================================================
    # =========  BAYESIAN MULTIMODEL COMPARISON ===========
    # =====================================================
    # ----- Define the discrepancy model -------
    sigma = np.ones(15) * np.array(sigma).flatten()
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = pd.DataFrame(sigma**2, columns=['Z'])

    # ----- Define the options model -------
    meta_models = {
        "linear": L2_engine,
        "exponential": NL2_engine,
        "cosine": NL4_engine
        }

    # BME Bootstrap options
    opts_bootstrap = {
        "bootstrap": True,
        "n_samples": 10,#0,#0, # TODO: difference between this and the n_bootstrap set below?
        "Discrepancy": DiscrepancyOpts,
        "emulator": False,
        "plot_post_pred": False
        }

    # Run model comparison
    BayesOpts = BayesModelComparison(
        justifiability=True,
        n_bootstrap=10,#0,#00,
        #just_n_meas=2
        emulator = False
        )
    # For Debug
    output_dict = BayesOpts.model_comparison_all(
        meta_models,
        opts_bootstrap
        )
    BayesOpts.model_names = ['linear', 'exponential', 'cosine']
    BayesOpts.calc_justifiability_analysis(meta_models, opts_bootstrap)

    # Save the results
    with open('model_comparison_output_dict.pkl', 'wb') as output:
        joblib.dump(output_dict, output, 2)
