#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov 20 14:48:43 2019

@author: farid
"""
import numpy as np
import scipy.stats as stats
import matplotlib.pyplot as plt
import scipy.stats as st
import seaborn as sns

def AnalyticalFunction(xx, t=None):
    """
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % Analytical Non-Gaussian Function
    %
    % Authors: Farid Mohammadi, University of Stuttgart
    %          Sergey Oladyshkin, University of Stuttgart
    % Questions/Comments: Please email Farid Mohammadi at:
    %   farid.mohammadi@iws.uni-stuttgart.de
    %
    % Copyright 2019. Farid Mohammadi, University of Stuttgart.
    %
    % THERE IS NO WARRANTY, EXPRESS OR IMPLIED. WE DO NOT ASSUME ANY LIABILITY
    % FOR THE USE OF THIS SOFTWARE.  If software is modified to produce
    % derivative works, such modified software should be clearly marked.
    % Additionally, this program is free software; you can redistribute it 
    % and/or modify it under the terms of the GNU General Public License as 
    % published by the Free Software Foundation; version 2.0 of the License. 
    % Accordingly, this program is distributed in the hope that it will be 
    % useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
    % of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
    % General Public License for more details.
    %
    % For function details and reference information, see:
    % https://doi.org/10.3390/e21111081
    %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %
    % OUTPUT AND INPUTS:
    %
    % Output = row vector of time vectors (s, t)
    %     Its structure is:
    %     y(t_1), y(t_2), ..., y(t_dt)
    % xx = [x1, x2, ..., xn]
    %       xn ~ Uinform(-5, 5)
    % t = vector of times (optional), with default value
    %     ( k − 1 ) /9 and k = 1,..., 10
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    """
    nParamSets, nParams = xx.shape
    
    if t is None: t = np.arange(0, 10, 1.) / 9
    
    term1 = (xx[:,0]**2 + xx[:,1] - 1)**2
    
    term2 = xx[:,0]**2
    
    term3 = 0.1 * xx[:,0] * np.exp(xx[:,1])
    
    term5 = 0
    if nParams > 2:
        for i in range(2, nParams):
            term5 = term5 + xx[:,i]**3/i
    
    const = term1 + term2 + term3 + 1 + term5
    
    # Compute time dependent term
    term4 = np.zeros((nParamSets,len(t)))
    for idx in range(nParamSets):
        term4[idx] = -2 * xx[idx,0] * np.sqrt(0.5*t)
    
    Output = term4 + np.repeat(const[:,None], len(t), axis=1)
    
    return np.vstack((t, Output))

if __name__ == "__main__":

    MCSize = 10000 #1000000
    ndim = 10
    sigma = 2
    
    # -------------------------------------------------------------------------
    # ----------------------- Synthetic data generation -----------------------
    # -------------------------------------------------------------------------
    t = np.arange(0, 10, 1.) / 9
    
    MAP = np.zeros((1, ndim))
    synthethicData = AnalyticalFunction(MAP, t=t)
    
    # -------------------------------------------------------------------------
    # ---------------------- Generate Prior distribution ----------------------
    # -------------------------------------------------------------------------
    
    xx = np.zeros((MCSize, ndim))
    
    params = (-5,5)
    
    for idxDim in range(ndim):
        lower, upper = params
        xx[:,idxDim] = stats.uniform(loc=lower, scale=upper-lower).rvs(size=MCSize)
    
    # -------------------------------------------------------------------------
    # ------------- BME and Kullback-Leibler Divergence -----------------------
    # -------------------------------------------------------------------------
    Outputs = AnalyticalFunction(xx, t=t)
    
    cov_matrix = np.diag(np.repeat(sigma**2, synthethicData.shape[1]))
    
    Likelihoods = st.multivariate_normal.pdf(Outputs[1:], mean=synthethicData[1], cov=cov_matrix)
    
    sns.kdeplot(np.log(Likelihoods[Likelihoods>0]), shade=True, color="g", label='Ref. Likelihood')
    
    normLikelihood = Likelihoods / np.nanmax(Likelihoods)
    # Random numbers between 0 and 1
    unif = np.random.rand(1, MCSize)[0]
    
    # Reject the poorly performed prior
    accepted = normLikelihood >= unif
    
    # Prior-based estimation of BME
    logBME = np.log(np.nanmean(Likelihoods))
    print('\nThe Naive MC-Estimation of BME is %.5f.'%(logBME))
    
    # Posterior-based expectation of likelihoods
    postExpLikelihoods = np.mean(np.log(Likelihoods[accepted]))
    
    # Calculate Kullback-Leibler Divergence
    KLD = postExpLikelihoods - logBME
    print("The Kullback-Leibler divergence estimation is %.5f."%KLD)
    
    # -------------------------------------------------------------------------
    # ----------------- Save the arrays as .npy files -------------------------
    # -------------------------------------------------------------------------
    if MCSize > 500000:
        np.save("data/refBME_KLD_"+str(ndim)+".npy", (logBME, KLD))
        np.save("data/mean_"+str(ndim)+".npy", np.mean(Outputs[1:],axis=0))
        np.save("data/std_"+str(ndim)+".npy", np.std(Outputs[1:],axis=0))
    else:
        np.save("data/Prior_"+str(ndim)+".npy", xx)
        np.save("data/origModelOutput_"+str(ndim)+".npy", Outputs[1:])
        np.save("data/validLikelihoods_"+str(ndim)+".npy", Likelihoods)
