#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 13 09:53:11 2020

@author: farid
"""
import sys, os
import numpy as np
import scipy.stats as stats
from itertools import cycle
import pandas as pd
try:
    import cPickle as pickle
except ModuleNotFoundError:
    import pickle
from sklearn import linear_model as lm
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
plt.rcParams.update({'font.size': 24})
plt.rc('figure', figsize = (24, 16))
plt.rc('font', family='serif', serif='Arial')
plt.rc('axes', grid = True)
plt.rc('text', usetex=True)
plt.rc('xtick', labelsize=24)
plt.rc('ytick', labelsize=24)
plt.rc('axes', labelsize=24)
plt.rc('axes', linewidth=2)
plt.rc('axes', grid=True)
plt.rc('grid', linestyle="-")
plt.rc('savefig', dpi=2000)

import matplotlib
matplotlib.use('agg')
# Local
sys.path.insert(0,'./../bayesian-validation/BayesValidRox/')

# Batch script
# sys.path.insert(0,'./../../bayesian-validation/BayesValidRox/')

from PyLink.FuncForwardModel import FuncForwardModel
from surrogate_models.Input import Input
from surrogate_models.surrogate_models import Metamodel
from PostProcessing.PostProcessing import PostProcessing


# Number of parameters
ndim = 10 # 2, 10

#=====================================================
#=============   COMPUTATIONAL MODEL  ================
#=====================================================
Model = FuncForwardModel()

Model.Type = 'Function'
Model.pyFile = 'AnalyticalFunction'
Model.Name = 'AnalyticFunc'

Model.Output.Names = ['Z']
OutputNames = ['Z']

# For Bayesian inversion synthetic data with X=[0,0]
Model.Observations['Time [s]'] = np.arange(0, 10, 1.) / 9
Model.Observations['Z'] = np.repeat([2],10)

# For Checking with the MonteCarlo refrence
Model.MCReference['Time [s]'] = np.arange(0, 10, 1.) / 9
Model.MCReference['mean'] = np.load("../data/mean_"+str(ndim)+".npy")
Model.MCReference['std'] = np.load("../data/std_"+str(ndim)+".npy")

#=====================================================
#=========   PROBABILISTIC INPUT MODEL  ==============
#=====================================================
# Define the uncertain parameters with their mean and 
# standard deviation
Inputs = Input()

for i in range(ndim):
    Inputs.addMarginals()
    Inputs.Marginals[i].Name = '$X_{%s}$'%(i+1)
    Inputs.Marginals[i].DistType = 'unif'
    Inputs.Marginals[i].Parameters =  [-5, 5]

# arbitrary polynomial chaos
# inputParams = np.load('../data/InputParameters_{}.npy'.format(ndim))
# for i in range(ndim):
#     Inputs.addMarginals()
#     Inputs.Marginals[i].Name = '$X_{%s}$'%(i+1)
#     Inputs.Marginals[i].InputValues = inputParams[:,i]

#=====================================================
#==========  DEFINITION OF THE METAMODEL  ============
#=====================================================    
MetaModelOpts = Metamodel(Inputs)

# Select if you want to preserve the spatial/temporal depencencies
MetaModelOpts.DimRedMethod = 'PCA'
MetaModelOpts.varPCAThreshold = 99.999

# Select your metamodel method
# 1) PCE (Polynomial Chaos Expansion) 2) GPE (Gaussian Process Emulator)
# 3) PCEKriging (PCE + GPE)
MetaModelOpts.metaModel = 'PCEKriging'

# ------------------------------------------------
# ------------- PCE Specification ----------------
# ------------------------------------------------
# Select the sparse least-square minimization method for 
# the PCE coefficients calculation:
# 1)OLS: Ordinary Least Square  2)BRR: Bayesian Ridge Regression 
# 3)LARS: Least angle regression  4)ARD: Bayesian ARD Regression
# 5)FastARD: Fast Bayesian ARD Regression
MetaModelOpts.RegMethod = 'FastARD'

# Specify the max degree to be compared by the adaptive algorithm:
# The degree with the lowest Leave-One-Out cross-validation (LOO)
# error (or the highest score=1-LOO)estimator is chosen as the final 
# metamodel.
MetaModelOpts.MinPceDegree = 6
MetaModelOpts.MaxPceDegree = 6
# q-quasi-norm 0<q<1 (default=1)
MetaModelOpts.q = 1.0 if ndim<5 else 0.75

# ------------------------------------------------
# ------ Experimental Design Configuration -------
# ------------------------------------------------
# Generate an experimental design of size NrExpDesign based on a latin 
# hypercube sampling of the input model or user-defined values of X and/or Y:
MetaModelOpts.addExpDesign()

# One-shot (normal) or Sequential Adaptive (sequential) Design
MetaModelOpts.ExpDesign.Method = 'normal'
MetaModelOpts.ExpDesign.NrSamples = 200

# Sampling methods
# 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley 6) chebyshev(FT) 
# 7) korobov 8) grid(FT) 9) nested_grid(FT) 10)user
MetaModelOpts.ExpDesign.SamplingMethod = 'random'

# >>>>>>>>>>>>>>>>>>>>>> Build Surrogate <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
# Adaptive sparse arbitrary polynomial chaos expansion
MetaModelOpts.index = 181
PCEModel = MetaModelOpts.create_metamodel(Model)

# PCEModel = BayesObj.PCEModel
# Extract slicing index
# Extract the experimental design
EDX = PCEModel.ExpDesign.X
index = PCEModel.index
EDYDict = PCEModel.ExpDesign.Y
JDist = PCEModel.ExpDesign.JDist
X_train = PCEModel.ExpDesign.X
EDY = PCEModel.ExpDesign.Y
#=====================================================
#=========  POST PROCESSING OF METAMODELS  ===========
#=====================================================
PostPCE = PostProcessing(PCEModel)

# Compute the moments and compare with the Monte-Carlo reference
PostPCE.plotMoments()

# Plot the sobol indices
sobol_cell, total_sobol = PostPCE.sobolIndicesPCE()


#=====================================================
#============  METAMODELS  WITH CHAOSPY  =============
#=====================================================
import chaospy
from surrogate_models.RegressionFastARD import RegressionFastARD

def PCATransformation(Output):
    
    # Transform via Principal Component Analysis
    from sklearn.decomposition import PCA as sklearnPCA

    n_samples, n_features = Output.shape
    covar_matrix = sklearnPCA(n_components=None, svd_solver='full')
    covar_matrix.fit(Output)
    var = np.cumsum(np.round(covar_matrix.explained_variance_ratio_, decimals=5)*100)
    try:
        selected_n_components = np.where(var>=99.0)[0][0] + 1
    except:
        selected_n_components = min(n_samples, n_features)
    
    nComponents = min(n_samples, n_features, selected_n_components)
    
    pca = sklearnPCA(n_components=nComponents, svd_solver='full', whiten=False)
    OutputMatrix = pca.fit_transform(Output)
    
    return pca, OutputMatrix

def fit_regression(methodConfig, X_train,Y_train,pca=False):
    
    from sklearn.multioutput import MultiOutputRegressor
    
    if pca:
        PCA, Y_train = PCATransformation(Y_train)
    else:
        PCA = None
        
    deg = methodConfig['deg']
    distribution = methodConfig['distribution']
    q = methodConfig['q']
    regmethod = methodConfig['regmethod'] 
    # Generate Orthogonal Polynomial Expansion:
    orthogonal_expansion = chaospy.orth_ttr(deg, distribution, cross_truncation=q)
    
    PCEModel={}
    PCEModel['config'] = methodConfig
    newPsi = orthogonal_expansion(*X_train.T).T

    if regmethod =='FastARD':
        PCEModel = MultiOutputRegressor(RegressionFastARD(fit_intercept=False)).fit(newPsi, Y_train)
    else:
        PCEModel = MultiOutputRegressor(lm.BayesianRidge(fit_intercept=False)).fit(newPsi, Y_train)

    # for idx in range(Y_train.shape[1]):
    #     clf_poly = RegressionFastARD(fit_intercept=False)
        
    #     PCEModel[idx] = clf_poly.fit(newPsi, Y_train[:,idx])
    
    return PCA, PCEModel

def approx_model(model, X,pca=None):
    
    global methodConfig
     
    y_Hat, y_Std = [], []
    deg = methodConfig['deg']
    distribution = methodConfig['distribution']
    q = methodConfig['q']
    
    # Generate Orthogonal Polynomial Expansion:
    orthogonal_expansion = chaospy.orth_ttr(deg, distribution, cross_truncation=q)
    
    psi = orthogonal_expansion(*X.T).T
    
    y_Hat = model.predict(psi).T
    
    # for idx in list(model.keys())[1:]:
    #     y_hat, y_std = model[idx].predict(psi, return_std=True)
    #     y_Hat.append(y_hat)
    #     y_Std.append(y_std)
    
    if pca is None:
       return np.array(y_Hat).T
    else:
        return pca.inverse_transform(np.array(y_Hat).T)# , np.array(y_Std).T

# Select a Distributions:
distribution = JDist #chaospy.J(*[chaospy.Uniform(-5.0,5.0)]*ndim)

methodConfig = dict()
y_chaos = dict()
methodConfig['regmethod'] = 'FastARD' #'FastARD'
methodConfig['deg'] = PCEModel.MaxPceDegree
methodConfig['distribution'] = distribution
methodConfig['q'] = MetaModelOpts.q


# Fit regression model
PCA = True
for key in OutputNames:
    
    pca, PCEModel_chaospy = fit_regression(methodConfig, X_train,EDY[key],pca=PCA)
    y_chaos[key] = approx_model(PCEModel_chaospy, X_train,pca=pca)


#=====================================================
#=================  Visualization  ===================
#=====================================================
# List of markers and colors
Color = ['k','b', 'g', 'r']
Marker = 'x'

Time = EDY['x_values']
      
nStes = 1
    
# Run pcemodel
y_hat, y_std = PCEModel.eval_metamodel(samples=EDX[:nStes]) #(nsamples=100, samples=EDX, name=case)

# create a PdfPages object
pdf = PdfPages('./PCE.pdf')
fig = plt.figure()
    
for key in OutputNames:
    
    fig, ax = plt.subplots()
    
    
    # Plot ExpDesign
    for i, output in enumerate(EDY[key][:nStes]):
        plt.plot(Time,output, alpha=0.35)
    
    
    # Plot PCE with BayesValidRox
    for i, output in enumerate(y_hat[key]):
        plt.plot(Time,output, ls='-', marker='*')
        
        
    # Plot PCE with Chaospy
    for i, output in enumerate(y_chaos[key][:nStes]):
        plt.plot(Time,output, ls='-', marker='*')
    
    plt.legend(loc='best', frameon=True)
    plt.xlabel('Time [s]')            
    plt.ylabel(key)
    
    # save the current figure
    pdf.savefig(fig, bbox_inches='tight')
        
    # Destroy the current plot
    plt.clf()

pdf.close() 


