#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This test shows a surrogate-assisted Bayesian calibration of a time dependent
    pollution function. Here, the noise will be jointly inferred with the input
    parameters.

Author: Farid Mohammadi, M.Sc.
E-Mail: farid.mohammadi@iws.uni-stuttgart.de
Department of Hydromechanics and Modelling of Hydrosystems (LH2)
Institute for Modelling Hydraulic and Environmental Systems (IWS), University
of Stuttgart, www.iws.uni-stuttgart.de/lh2/
Pfaffenwaldring 61
70569 Stuttgart

Created on Fri Aug 9 2019

"""

import numpy as np
import pandas as pd
import joblib

# import bayesvalidrox
# Add BayesValidRox path
import sys
sys.path.append("../../src/bayesvalidrox/")

from pylink.pylink import PyLinkForwardModel
from surrogate_models.inputs import Input
from surrogate_models.surrogate_models import MetaModel
from post_processing.post_processing import PostProcessing
from bayes_inference.bayes_inference import BayesInference
from bayes_inference.discrepancy import Discrepancy

if __name__ == "__main__":

    # =====================================================
    # =============   COMPUTATIONAL MODEL  ================
    # =====================================================
    Model = PyLinkForwardModel()

    # Define model options
    Model.link_type = 'Function'
    Model.py_file = 'pollution'
    Model.name = 'pollution'

    Model.Output.names = ['C']

    # Synthetic data for Bayesian inversion.
    # MAP = (10, 0.07, 1.0, 30.16)
    Model.observations = {}
    Model.observations['Time [s]'] = [40+5*i for i in range(21)]
    Model.observations['C'] = np.load("data/MeasuredData.npy")

    # For Checking with the MonteCarlo refrence
    Model.mc_reference = {}
    Model.mc_reference['Time [s]'] = [40+5*i for i in range(21)]
    Model.mc_reference['mean'] = np.load("data/mean.npy")
    Model.mc_reference['std'] = np.load("data/std.npy")

    # =====================================================
    # =========   PROBABILISTIC INPUT MODEL  ==============
    # =====================================================
    # Define the uncertain parameters with their mean and
    # standard deviation
    Inputs = Input()

    # mass of pollutant spilled at each location
    Inputs.add_marginals()
    Inputs.Marginals[0].name = 'M'
    Inputs.Marginals[0].dist_type = 'uniform'
    Inputs.Marginals[0].parameters = [7, 13]

    # diffusion rate in the channel
    Inputs.add_marginals()
    Inputs.Marginals[1].name = 'D'
    Inputs.Marginals[1].dist_type = 'uniform'
    Inputs.Marginals[1].parameters = [0.02, 0.12]

    # location of the second spill
    Inputs.add_marginals()
    Inputs.Marginals[2].name = 'L'
    Inputs.Marginals[2].dist_type = 'uniform'
    Inputs.Marginals[2].parameters = [0.01, 3]

    # time of the second spill
    Inputs.add_marginals()
    Inputs.Marginals[3].name = '$\\tau$'
    Inputs.Marginals[3].dist_type = 'uniform'
    Inputs.Marginals[3].parameters = [30.01, 30.295]

    # =====================================================
    # ======  POLYNOMIAL CHAOS EXPANSION METAMODELS  ======
    # =====================================================
    MetaModelOpts = MetaModel(Inputs, Model)

    # Select if you want to preserve the spatial/temporal depencencies
    # MetaModelOpts.dim_red_method = 'PCA'
    # MetaModelOpts.var_pca_threshold = 99.999
    # MetaModelOpts.n_pca_components = 12

    # Select your metamodel method
    # 1) PCE (Polynomial Chaos Expansion) 2) aPCE (arbitrary PCE)
    # 3) GPE (Gaussian Process Emulator)
    MetaModelOpts.meta_model_type = 'aPCE'

    # ------------------------------------------------
    # ------------- PCE Specification ----------------
    # ------------------------------------------------
    # Select the sparse least-square minimization method for
    # the PCE coefficients calculation:
    # 1)OLS: Ordinary Least Square  2)BRR: Bayesian Ridge Regression
    # 3)LARS: Least angle regression  4)ARD: Bayesian ARD Regression
    # 5)FastARD: Fast Bayesian ARD Regression
    # 6)BCS: Bayesian Compressive Sensing
    # 7)OMP: Orthogonal Matching Pursuit
    # 8)VBL: Variational Bayesian Learning
    # 9)EBL: Emperical Bayesian Learning
    MetaModelOpts.pce_reg_method = 'BCS'

    # Bootstraping
    # 1) normal 2) fast
    MetaModelOpts.bootstrap_method = 'fast'
    MetaModelOpts.n_bootstrap_itrs = 100

    # Specify the max degree to be compared by the adaptive algorithm:
    # The degree with the lowest Leave-One-Out cross-validation (LOO)
    # error (or the highest score=1-LOO)estimator is chosen as the final
    # metamodel. pce_deg accepts degree as a scalar or a range.
    MetaModelOpts.pce_deg = 8

    # q-quasi-norm 0<q<1 (default=1)
    # MetaModelOpts.pce_q_norm = 0.75

    # Print summary of the regression results
    # MetaModelOpts.verbose = True

    # ------ Experimental Design --------
    MetaModelOpts.add_ExpDesign()

    # One-shot (normal) or Sequential Adaptive (sequential) Design
    MetaModelOpts.ExpDesign.method = 'normal'
    MetaModelOpts.ExpDesign.n_init_samples = 150

    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley 6) korobov
    # 7) chebyshev(FT) 8) grid(FT) 9) nested_grid(FT) 10)user
    MetaModelOpts.ExpDesign.sampling_method = 'latin_hypercube'

    # Provide the experimental design object with a hdf5 file
    # MetaModelOpts.ExpDesign.hdf5_file = 'ExpDesign_AnalyticFunc.hdf5'

    # Sequential experimental design (needed only for sequential ExpDesign)
    MetaModelOpts.ExpDesign.n_new_samples = 1
    MetaModelOpts.ExpDesign.n_max_samples = 15  # 150
    MetaModelOpts.ExpDesign.mod_LOO_threshold = 1e-16

    # Defining the measurement error, if it's known a priori
    obsData = pd.DataFrame(0.5*np.ones(21), columns=Model.Output.names)
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = obsData ** 2
    MetaModelOpts.Discrepancy = DiscrepancyOpts

    # Plot the posterior snapshots for SeqDesign
    MetaModelOpts.ExpDesign.post_snapshot = False
    MetaModelOpts.ExpDesign.step_snapshot = 1
    MetaModelOpts.ExpDesign.max_a_post = (10, 0.07, 1.0, 30.16)

    # ------------------------------------------------
    # ------- Sequential Design configuration --------
    # ------------------------------------------------
    # MetaModelOpts.adapt_verbose = True

    # -------- Tradeoff scheme ------
    # 1) 'None' 2) 'equal' 3)'epsilon-decreasing' 4) 'adaptive'
    MetaModelOpts.ExpDesign.tradeoff_scheme = 'epsilon-decreasing'
    # MetaModelOpts.ExpDesign.n_replication = 20

    # -------- Exploration ------
    # 1)'Voronoi' 2)'random' 3)'latin_hypercube' 4)'LOOCV' 5)'dual annealing'
    MetaModelOpts.ExpDesign.explore_method = 'random'

    # Use when 'dual annealing' chosen
    MetaModelOpts.ExpDesign.max_func_itr = 200

    # Use when 'Voronoi' or 'random' or 'latin_hypercube' chosen
    MetaModelOpts.ExpDesign.n_canddidate = 5000
    MetaModelOpts.ExpDesign.n_cand_groups = 8

    # -------- Exploitation ------
    # 1)'BayesOptDesign' 2)'BayesActDesign' 3)'VarOptDesign' 4)'alphabetic'
    # 5)'Space-filling'
    MetaModelOpts.ExpDesign.exploit_method = 'VarOptDesign'

    # BayesOptDesign -> when data is available
    # 1)DKL (Kullback-Leibler Divergence) 2)DPP (D-Posterior-percision)
    # 3)APP (A-Posterior-percision)  # ['DKL', 'BME', 'infEntropy']
    # MetaModelOpts.ExpDesign.util_func = ['DKL', 'BME']

    # VarBasedOptDesign -> when data is not available
    # Only with Vornoi >>> 1)Entropy 2)EIGF, 3)LOOCV
    MetaModelOpts.ExpDesign.util_func = 'Entropy'

    # alphabetic
    # 1)D-Opt (D-Optimality) 2)A-Opt (A-Optimality)
    # 3)K-Opt (K-Optimality)
    # MetaModelOpts.ExpDesign.util_func = 'D-Opt'

    # For colculation of validation error for SeqDesign
    MetaModelOpts.valid_samples = np.load("data/validSet.npy")
    MetaModelOpts.valid_model_runs = {'C': np.load("data/validSetOutput.npy")}
    MetaModelOpts.valid_likelihoods = np.load("data/validLikelihoods.npy")

    # >>>>>>>>>>>>>>>>>>>>>> Build Surrogate <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    # Adaptive sparse arbitrary polynomial chaos expansion
    PCEModel = MetaModelOpts.create_metamodel()

    # =====================================================
    # =========  POST PROCESSING OF METAMODELS  ===========
    # =====================================================
    PostPCE = PostProcessing(PCEModel)

    # Plot to check validation visually.
    PostPCE.valid_metamodel(n_samples=3)

    # Compute the moments and compare with the Monte-Carlo reference
    PostPCE.plot_moments()

    # Plot the evolution of the KLD,BME, and Modified LOOCV error
    if MetaModelOpts.ExpDesign.method == 'sequential':
        refBME_KLD = np.load("data/refBME_KLD.npy")
        PostPCE.plot_seq_design_diagnostics(refBME_KLD)

    # Plot the sobol indices
    if MetaModelOpts.meta_model_type != 'GPE':
        total_sobol = PostPCE.sobol_indices()

    # Compute and print RMSE error
    valid_out_dict = dict()
    valid_out_dict['Time [s]'] = [40+5*i for i in range(21)]
    valid_out_dict['C'] = np.load("data/validSetOutput.npy")
    PostPCE.check_accuracy(samples=MetaModelOpts.valid_samples,
                           outputs=valid_out_dict)

    # =====================================================
    # ========  Bayesian inference with Emulator ==========
    # =====================================================
    BayesOpts = BayesInference(PCEModel)
    BayesOpts.emulator = True

    # Select the inference method
    BayesOpts.inference_method = "MCMC"
    # Set the MCMC parameters
    BayesOpts.mcmc_params = {
        'n_steps': 1e5,
        'n_walkers': 30,
        'multiprocessing': False,
        'verbose': False
        }

    BayesOpts.plot_post_pred = True

    # ----- Define the discrepancy model -------
    # Let us infer the sigma2
    DiscOutputOpts = Input()
    # # OutputName = 'Z'
    DiscOutputOpts.add_marginals()
    DiscOutputOpts.Marginals[0].name = '$\\sigma^2_{\\epsilon}$'
    DiscOutputOpts.Marginals[0].dist_type = 'uniform'
    DiscOutputOpts.Marginals[0].parameters = [0.0, 5.0]
    BayesOpts.Discrepancy = Discrepancy(DiscOutputOpts)

    Bayes_PCE = BayesOpts.create_inference()

    # Save class objects
    with open(f'Bayes_{Model.name}.pkl', 'wb') as output:
        joblib.dump(PCEModel, output, 2)
        joblib.dump(PostPCE, output, 2)
        joblib.dump(Bayes_PCE, output, 2)
