# -*- coding: utf-8 -*-
"""
Example and test for using UM-Bridge with bayesvalidrox.
This example uses the `testmodel` and sets the model type explicitly as 
UM-Bridge for the pylink model.

An example for using UM-Bridge with an explicitly defined model can be found in 
`example_umbridge_tsunamimodel.py`

"""

import joblib
import numpy as np 
import umbridge
import bayesvalidrox as bv

from bayesvalidrox.pylink.pylink import PyLinkForwardModel
from bayesvalidrox.surrogate_models.inputs import Input
from bayesvalidrox.surrogate_models.exp_designs import ExpDesigns
from bayesvalidrox.surrogate_models.surrogate_models import MetaModel
from bayesvalidrox.post_processing.post_processing import PostProcessing
from bayesvalidrox.surrogate_models.engine import Engine


if __name__ == '__main__':
    
    # This model has 2 inputs and four outputs
    n_prior_sample = 1000
    priors = bv.Input()
    priors.add_marginals()
    priors.Marginals[0].name = 'x'
    priors.Marginals[0].input_data = np.random.uniform(50,150,n_prior_sample)  
    
    # Define the model directly via Pylink
    model = PyLinkForwardModel()
    model.link_type = 'umbridge'
    model.host = 'http://testmodel.linusseelinger.de'
    model.modelparams = {}
    model.name = 'testmodel'
    model.Output.names = ['y']
    model.x_values = np.array([0])
    
    # Create the surrogate
    surrogate_opts = MetaModel(priors)
    
    # Select the surrogate type and properties
    surrogate_opts.meta_model_type = 'aPCE'
    surrogate_opts.pce_reg_method = 'FastARD'
    surrogate_opts.pce_deg = np.arange(1, 5)
    surrogate_opts.pce_q_norm = 0.4

    # Define the experimental design and sampling methods
    ExpDesign = ExpDesigns(priors)     
    ExpDesign.n_init_samples = 50
    ExpDesign.sampling_method = 'latin-hypercube'
    
    # Start and run the engine
    engine = Engine(surrogate_opts, model, ExpDesign)
    engine.start_engine()
    engine.train_normal(parallel = False)
    print('Surrogate completed')
    print('')

    # Save surrogate
    with open('testmodel.pk1', 'wb') as output:
        joblib.dump(engine.MetaModel, output, 2)
        
    # Post processing
    L2_PostPCE = PostProcessing(engine)
    L2_PostPCE.plot_moments(plot_type='line')
    
    # Plot to check validation visually.
    L2_PostPCE.valid_metamodel(n_samples=1)
    
    # Compute and print RMSE error
    L2_PostPCE.check_accuracy(n_samples=300)
    total_sobol = L2_PostPCE.sobol_indices()
