# -*- coding: utf-8 -*-
"""
Example and test for using UM-Bridge with bayesvalidrox.
Example model is the tsunami model described here:
    https://um-bridge-benchmarks.readthedocs.io/en/docs/inverse-benchmarks/exahype-tsunami.html
It needs to run separately while starting this example. 
    docker run -it -p 4242:4242 -v ~/tsunami_output:/output linusseelinger/model-exahype-tsunami
    
This example uses models that are exmplicitly defined as functions outside of
PyLink. An example for the link_type 'umbdridge' can be found in 
`example_umbridge_testmodel.py`.

"""

import copy
import joblib
import numpy as np 
import pandas as pd
import umbridge

import bayesvalidrox as bv
from tsunami_model import tsunami_model

if __name__ == '__main__':
    # This model has 2 inputs and four outputs
    n_prior_sample = 1000
    priors = bv.Input()
    priors.add_marginals()
    priors.Marginals[0].name = 'x'
    priors.Marginals[0].input_data = np.random.uniform(50,150,n_prior_sample)  
    priors.add_marginals()
    priors.Marginals[1].name = 'y'
    priors.Marginals[1].input_data = np.random.uniform(50,150,n_prior_sample)
    
    # Define the model - level 0
    model0 = bv.PyLinkForwardModel()
    model0.link_type = 'function'
    model0.py_file = 'tsunami_model'
    model0.name = 'tsunami_model'
    model0.Output.names = ['T1', 'T2', 'H1', 'H2']
    #model.observations = data_dict1_times
    
    # Define the model - level 1
    model1 = bv.PyLinkForwardModel()
    model1.link_type = 'function'
    model1.py_file = 'tsunami_model1'
    model1.name = 'tsunami_model1'
    model1.Output.names = ['T1', 'T2', 'H1', 'H2']
    #model.observations = data_dict1_times
    
    # Create the surrogate
    surrogate_opts0 = bv.MetaModel(priors, model0)
    
    # Select your metamodel method
    surrogate_opts0.meta_model_type = 'aPCE'
    surrogate_opts0.pce_reg_method = 'FastARD'
    surrogate_opts0.pce_deg = np.arange(1, 5)
    surrogate_opts0.pce_q_norm = 0.4#1.0

    # Define ExpDesign - this is the same for both models       
    ExpDesign0 = bv.ExpDesigns(priors)
    ExpDesign0.method = 'normal'
    ExpDesign0.n_init_samples = 50
    ExpDesign0.sampling_method = 'latin-hypercube'
    ExpDesign1 = copy.deepcopy(ExpDesign0)
    
    # Start and run the engine
    engine0 = bv.Engine(surrogate_opts0, model0, ExpDesign0)
    engine0.start_engine()
    engine0.train_normal(parallel = False)
    print('Surrogate 0 completed')
    print('')
    
    # Create the surrogate
    surrogate_opts1 = bv.MetaModel(priors, model1)
    
    # Select your metamodel method
    surrogate_opts1.meta_model_type = 'aPCE'
    surrogate_opts1.pce_reg_method = 'FastARD'
    surrogate_opts1.pce_deg = np.arange(1, 5)
    surrogate_opts1.pce_q_norm = 0.4#1.0

    # Start and run the engine
    engine1 = bv.Engine(surrogate_opts1, model1, ExpDesign1)
    engine1.start_engine()
    engine1.train_normal(parallel = False)
    print('Surrogate 1 completed')
    print('')

    # Save surrogate
    with open('tsunami0.pk1', 'wb') as output:
        joblib.dump(engine0.MetaModel, output, 2)
        
    # Save surrogate
    with open('tsunami1.pk1', 'wb') as output:
        joblib.dump(engine1.MetaModel, output, 2)
    
    # Post processing on model 1
    L2_PostPCE = bv.PostProcessing(engine1)
    L2_PostPCE.plot_moments(plot_type='line')
    # Plot to check validation visually.
    L2_PostPCE.valid_metamodel(n_samples=1)
    
    # Compute and print RMSE error
    L2_PostPCE.check_accuracy(n_samples=300)
    total_sobol = L2_PostPCE.sobol_indices()

    # Get reference evaluation as 'true data'
    true_data = tsunami_model(np.array([[90.0,60.0]]))
    model0.observations = true_data
    model1.observations = true_data
    true_data_nox = copy.deepcopy(true_data)
    true_data_nox.pop('x_values')
    
    # Direct surrogate evaluation
    mean, std = engine1.MetaModel.eval_metamodel(np.array([[90.0,60.0]]))
    
    # Bayesian Inference
    # Set uncertainty
    obsData = pd.DataFrame(true_data_nox, columns=model0.Output.names)
    DiscrepancyOpts = bv.Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = (obsData*0.15)**2
    
    # Parameter estimation / single model validation via TOM for model 1
    BayesOpts = bv.BayesInference(engine1)
    BayesOpts.emulator= True
    BayesOpts.plot_post_pred = True
    #BayesOpts.inference_method = 'rejection'
    BayesOpts.Discrepancy = DiscrepancyOpts
    BayesOpts.bootstrap = True
    BayesOpts.n_bootstrap_itrs = 10
    BayesOpts.bootstrap_noise = 0.05
    BayesOpts.out_dir = ''
    
    # Set the MCMC parameters
    import emcee
    BayesOpts.inference_method = "MCMC"
    BayesOpts.mcmc_params = {
        'n_steps': 1e3,
        'n_walkers': 30,
        'moves': emcee.moves.KDEMove(),
        'multiprocessing': False,
        'verbose': False
        }
     
    Bayes = BayesOpts.create_inference()
    
    # Pkl the inference results
    with open(f'Bayes_{model0.name}.pkl', 'wb') as output:
        joblib.dump(Bayes, output, 2)
            
    # Model Comparison
    # Set the models
    meta_models = {
        "M0": engine0,
        "M1": engine1,
        }

    # BME Bootstrap options
    opts_bootstrap = {
        "bootstrap": True,
        "n_samples": 1000,
        "Discrepancy": DiscrepancyOpts,
        "emulator": True,
        "plot_post_pred": False
        }

    # Run model comparison
    BayesOpts = bv.BayesModelComparison(
        justifiability=True,
        n_bootstarp=10,
        just_n_meas=2
        )
    output_dict = BayesOpts.create_model_comparison(
        meta_models,
        opts_bootstrap
        )
            
    # save model comparison results
    with open('ModelComparison.pkl', 'wb') as f:
        joblib.dump(output_dict, f)
    